//  Copyright (c) 2016-2021  Milos Jakubicek

#include <finlib/regpref.hh>
#include "subcorp.hh"
#include "keyword.hh"

bool check_string (string &str, vector<regexp_pattern*> pos_filters, vector<regexp_pattern*> neg_filters,
                   unordered_set<string> &blist, unordered_set<string> &wlist)
{
    for (auto it = neg_filters.begin(); it != neg_filters.end(); it++) {
        if ((*it)->match(str.c_str()))
            return false;
    }
    for (auto it = pos_filters.begin(); it != pos_filters.end(); it++) {
        if (!(*it)->match(str.c_str()))
            return false;
    }
    if (!blist.empty() && blist.find(str) != blist.end())
        return false;
    if (!wlist.empty() && wlist.find(str) == wlist.end())
        return false;
    return true;
}

class AllowMissingFrequency : public Frequency {
    Frequency *src = NULL;
public:
    AllowMissingFrequency (WordList *wl, const char *frqtype) {
        try {
            src = wl->get_stat(frqtype);
        } catch (FileAccessError&) {}
    }
    double freq(int id) {return src ? src->freq(id) : -1;}
    ~AllowMissingFrequency() {delete src;}
};

Keyword::Keyword (Corpus *c1, Corpus *c2, WordList *wl1, WordList *wl2, float N,
             unsigned maxlen, int minfreq, int maxfreq, unordered_set<string> &blacklist,
             unordered_set<string> &whitelist, const char *frqtype, vector<string> &addfreqs,
             vector<string> pos_regex_filters, vector<string> neg_regex_filters, FILE* progress)
            : curr (0), totalcount(0), totalfreq1(0), totalfreq2(0)
{
    vector<regexp_pattern*> pos_regpats;
    for (auto it = pos_regex_filters.begin(); it != pos_regex_filters.end(); it++) {
        if (!(*it).size())
            continue;
        regexp_pattern *re = new regexp_pattern((*it).c_str(), wl1->locale, wl1->encoding);
        if (re->compile())
            throw new CorpInfoNotFound("Invalid regular expression for Keyword");
        pos_regpats.push_back(re);
    }

    vector<regexp_pattern*> neg_regpats;
    for (auto it = neg_regex_filters.begin(); it != neg_regex_filters.end(); it++) {
        if (!(*it).size())
            continue;
        regexp_pattern *re = new regexp_pattern((*it).c_str(), wl1->locale, wl1->encoding);
        if (re->compile())
            throw new CorpInfoNotFound("Invalid regular expression for Keyword");
        neg_regpats.push_back(re);
    }

    strid_gen *it = NULL;
    if (c1->get_confpath() == c2->get_confpath())
        it = new strid_gen (wl1->dump_str());
    else
        it = new strid_gen (new WordListLeftJoin (wl1->attr_path, wl2->attr_path));

    const float c1size = c1->search_size();
    const float c2size = c2->search_size();
    const float wl1_maxid = wl1->id_range();
    Frequency *stat1 = wl1->get_stat(frqtype);
    Frequency *stat2 = wl2->get_stat(frqtype);
    vector<Frequency*> addfreqs1, addfreqs2;
    for (auto it = addfreqs.begin(); it != addfreqs.end(); it++) {
        addfreqs1.push_back(new AllowMissingFrequency(wl1, (*it).c_str()));
        addfreqs2.push_back(new AllowMissingFrequency(wl2, (*it).c_str()));
    }

    while (!it->end()) {
        string str;
        int id1, id2;
        it->next (str, id1, id2);
        if (progress && id1 % 100000 == 0)
            fprintf(progress, "\r%.2f %%", id1 / wl1_maxid * 100);
        NumOfPos f1 = stat1->freq(id1);
        NumOfPos f2;
        if (id2 == -1)
            f2 = 0;
        else
            f2 = stat2->freq(id2);
        if (minfreq && f1 < minfreq)
            continue;
        if (maxfreq && f1 > maxfreq)
            continue;
        totalcount++;
        totalfreq1 += f1;
        totalfreq2 += f2;
        float fpm1 = f1 * 1000000 / c1size;
        float fpm2 = f2 * 1000000 / c2size;
        double *freqs = new double[2*addfreqs.size() + 4];
        freqs[0] = f1; freqs[1] = f2;
        freqs[2] = fpm1; freqs[3] = fpm2;
        for (unsigned i = 0; i < addfreqs.size(); i++) {
            freqs[2*i+4] = addfreqs1[i]->freq(id1);
            freqs[2*i+5] = id2 == -1 ? 0 : addfreqs2[i]->freq(id2);
        }
        float score = (fpm1 + N) / (fpm2 + N);
        if (heap.size() < maxlen) {
            if (!check_string (str, pos_regpats, neg_regpats, blacklist, whitelist))
                continue;
            kwitem *k = new kwitem (id1, id2, score, str, freqs);
            heap.push_back (k);
            push_heap (heap.begin(), heap.end(), kwitem_cmp());
        } else if (heap.front()->score < score) {
            if (!check_string (str, pos_regpats, neg_regpats, blacklist, whitelist))
                continue;
            pop_heap (heap.begin(), heap.end(), kwitem_cmp());
            delete heap.back();
            heap.pop_back();
            kwitem *k = new kwitem (id1, id2, score, str, freqs);
            heap.push_back (k);
            push_heap (heap.begin(), heap.end(), kwitem_cmp());
        }
    }
    delete stat1; delete stat2;
    for (auto it = pos_regpats.begin(); it != pos_regpats.end(); it++)
        delete *it;
    for (auto it = neg_regpats.begin(); it != neg_regpats.end(); it++)
        delete *it;
    if (progress)
        fprintf(progress, "\r100 %%     \n");
    sort_heap (heap.begin(), heap.end(), kwitem_cmp());
}

// vim: ts=4 sw=4 sta et sts=4 si cindent tw=80:
